/*
 * Copyright (c) 2020 Mark Jamsek <mark@jamsek.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* 
 * CSP2348 Assignment 2: A Mini Team Project (Heap sort algorithm and analysis)
 * heap_sort.c
 * Purpose: For a given array of integers, implement the heap sort algorithm to
 * return the array sorted in ascending order as defined in Task 1 of the
 * assessment documentation.
 *
 * @authors Mark Jamsek
 * @version 0.1 2020-04-10
 * 
 * Coding Style: BSD KNF (Kernel Normal Form) <https://man.openbsd.org/style>
 * Coding Standard: ISO C18 <http://www.iso-9899.info/wiki/The_Standard#C18>
 *
 * Tested on macOS Catalina 10.15.{3,4}, and OpenBSD 6.{3,6}.
 *   
 * Unix: clang -Wall -Werror -O3 -g -o heaps heap_sort.c
 *   - run with ./heaps
 * 
 * main() routine initialises an array of N random integers with a distribution
 * of DIST_MIN to DIST_MAX before passing the array to heap_sort() for sorting.
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define N           20      /* Number of array elements for RNG to generate */
#define DIST_MIN    -100    /* Range minimum of random distribution */
#define DIST_MAX    100     /* Range maximum of random distribution */

void        heap_sort(int *, int);
void        _heapify(int *, int, int);
int         _nmax(int *, int, int, int, int);
void        randints(int *, int, int, int);
void        display_array(int *, int);

int
main(int argc, const char **argv)
{
        int     arr[] = {128, 16, 48, 10, 512, 2, -32, 8, 2048, 0, -1, 96, 21,
                            -88, 64, 256, 10, 0, 42, 13};
        int     n = sizeof arr / sizeof arr[0];
        int     r[N];

        /* Generate random numbers for array to be sorted */
        randints(r, DIST_MIN, DIST_MAX, N);

        printf("\n[-] unsorted array(s)\n\t");
        display_array(arr, n);
        display_array(r, N);
        
        /* Sort array(s) */
        heap_sort(arr, n);
        heap_sort(r, N);
        
        printf("\n[+] sorted array(s)\n\t");
        display_array(arr, n);
        display_array(r, N);

        printf("\n");
        
        return 0;
}

void
heap_sort(int *a, int n)
{
        /*
         * Step 1. Build heap: relocate max node to front
         */
        for (int i = n / 2 - 1; i >= 0; --i)
                _heapify(a, i, n);
        
        /* Traverse heap */
        for (int i = n - 1; i > 0; --i) {
                
                /*
                 * Step 2.1 Transpose root with tail: each iteration move max
                 * node to the end where it's in place & remove from the heap
                 */
                int tmp = a[i];
                a[i] = a[0];
                a[0] = tmp;
                
                /*
                 * Step 2.2 Reheap reduced heap
                 */
                _heapify(a, 0, i);
        }
}

void
_heapify(int *a, int node, int n)
{
        /* Find maximal node */
        int max = _nmax(a, n, node, node * 2 + 1, node * 2 + 2);

        /* If parent is not max, swap with child and continue down heap */
        if (node != max) {
                a[max] ^= a[node];
                a[node] ^= a[max];
                a[max] ^= a[node];
                node = max;
                _heapify(a, node, n);
        }
}

int
_nmax(int *a, int n, int parent, int l, int r)
{
        /* Initialise parent as max node */
        int     max = parent;
        
        /* Left child is greatest */
        if (l < n && a[l] > a[max])
                max = l;
        
        /* Right child is greatest */
        if (r < n && a[r] > a[max])
                max = r;
        
        return max;
}

void
randints(int *r, int min, int max, int n)
{
        srand(time(0));     /* Seed RNG */

        for (int i = 0; i < n; ++i)
                r[i] = (rand() % (max - min + 1)) + min;
}

void
display_array(int *a, int n)
{
        for (int i = 0; i < n; ++i)
                printf("%d%s", a[i], i == n - 1 ? "\n\t" : ", ");
}
